/////////////////////////////////////////////////////////////////////////////
/*
  Copyright 2001 Ronald S. Burkey

  This file is part of GutenMark.

  GutenMark is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GutenMark is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GutenMark; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	MarkBody.c
  Purpose:	This analyzes the body of the text.
  Mods:		08/31/01 RSB	Began.
  		09/03/01 RSB	N, S, E, and W added to honorifics (really
				street names, but works the same).
		09/04/01 RSB	Now replace '<' & '>' & '&', since they 
				can be mistaken for markup.  Add closing 
				markup on reaching end of file.
		09/06/01 RSB	Process paragraphs to try and deduce 
				whether to make them justified or ragged.	
		09/08/01 RSB	Lotsa cleanup and tweaking.	
		09/15/01 RSB	Now account for the ~~ italics delimiters.
				Also, the /italics/ format.
				Added more heuristics for detecting verse.
		11/02/01 RSB	Added GPL disclaimer and reformatted 
				somewhat for first web release.	
		11/09/01 RSB	Now convert all strings of dashes of
				length 2 or more to mdashes and ndashes.
				In other words, "--" converts to &mdash;
				"---" converts to &mdash;&ndash;
				"----" converts to &mdash;&mdash; and so on.
				These don't look good in html, but look
				great in postscript or pdf.  Also, the
				dash in " - " is converted to an mdash.	
		11/12/01 RSB	Added detection of the PG file-ender.
		11/13/01 RSB	Added smart quotes.
		11/23/01 RSB	Added ALL-CAPs and foreign italicizing.
		11/24/01 RSB	Centralized all of the markup-file output
				operations into the AddMarkup function.
				8-bit ASCII restoration.
		11/25/01 RSB	Resolved conflicts between 8-bit restoration
				and ALL-CAPS/foreign markup.										
  
  Styles of italicization are indicated by the Italicizing variable, 
  as follows:
  	A	All caps.
	H	HTML -- like <I>hello</I>.
	_	Like _hello_.
	/	Like /hello/.
	~	Like ~~hello~~.
	>	Like <hello>.
  As might be imagined, there's some difficulty distinguishing between 
  the styles <I>hello</I> and <hello>.  To this end, a variable called 
  BracketItalicsCount is kept; initially, the program guesses which style 
  is which, but after enough of one style or the other is encountered, 
  as recorded by BracketItalicsCount, the style is locked in.
*/  
  
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "AutoMark.h"
  
// Number of lines prior to and following the current line to consider in
// the analysis.
#define PRE_OR_POST_LINES 10
#define BUFFERED_LINES (2*PRE_OR_POST_LINES+1)
  
// Number of headings which, once matched from the prefatory area, 
// force all further matches to come from there (or from special words).
#define PREFATORY_MATCH_CEILING 4
  
// We spell these honorifics in reverse to make the search easier.
// No need to add single-letter ones, because these are treated specially.
// I've also added various names that appear in older texts as 
// abbreviations, like Jas., Chas., Ed., etc.
static char *Honorifics[] =
  { 
"rM", "RM", "srM", "SRM", "rD", "RD", "srD", "SRD", "tS", "TS", "etS",

"ETS", 
"arF", "ARF", "rF", "RF", "emM", "EMM", "sM", "SM", 
"emdM", "EMDM", "ellM", "ELLM",
"snoM", "SNOM", 
"sahC", "SAHC", "saJ", "SAJ", "dE", "DE", "sirhC", "SIRHC", 
"oeG", "OEG", "derF",
"DERF", "mW", "MW" 
};


// Strings to auto-italicize.  Note that adding "PS." to the list has the
// happy side effect of not adding spaces to it in end-of-sentence detection.
static char *Italicized[] =
  { 
"Etc.", "etc.", "viz.", "ie.", "i.e.", "Ie.", "I.e.", "eg.", "e.g.",

"Eg.", "E.g.", "et al.", "et cetera", "n.b.", "N.b.", "nota bene", 
"Nota bene", /* "M.", */
"Ste.", "Mme.", "Mdme.", "Mlle.", "Mons.", 
  "PS." "PPS." 
};

#define NUM_AUTO_ITALICIZED (sizeof(Italicized)/sizeof(Italicized[0]))
static int LengthItalicized[NUM_AUTO_ITALICIZED];


//-----------------------------------------------------------------------
// Add a markup record to the markup file.  Returns 0 on success.
// This function plays a slightly more complex role than it would seem
// at first glance.  The markup file must be sorted in order of increasing
// offsets into the original etext, and in almost all cases the program 
// actually tries to add the markup in this order.  However, there is a
// small number of cases (right now, involving only italicizing foreign
// text) in which the program finds it necessary to insert a markup record
// a few records back.  Rather than applying some complex logic to track
// this, or sorting the (large) markup file at the end of the process,
// we take care of this completely within AddMarkup.  
//
// If called with Offset == -1, AddMarkup's internal buffer (if any) is 
// instead flushed to disk, so that the file can be safely closed.
//
// Returns 0 on success.

static long LastOffset = -1;	// Max offset used so far.
static int 
AddMarkup (AnalysisDataset * Dataset, long Offset, enum MarkupType Type,
	   
char Insert) 
{
  
if (Offset >= 0)
    
    {
      
MarkupRecord Record;
      
Record.Offset = Offset;
      
Record.Type = Type;
      
Record.Insert = Insert;
      
	// If the Offset of the new markup is less than the Offset of
	// previous markup, we need to backtrack in the output file,
	// so as to insert the record at the proper point.  The seeking
	// is inefficient, but we expect to do it very seldom, and for 
	// only a few record-positions.
	if (Offset < LastOffset)
	
	{
	  
long CurrentPos;
	  
int NumRecs;
	  
MarkupRecord CheckRecord, *TempRecords;
	  
NumRecs = 0;
	  
	  do
	    
	    {
	      
if (fseek 
		   (Dataset->MarkupFile, -sizeof (CheckRecord), SEEK_CUR))
		
return (1);
	      
if (1 !=
		   
fread (&CheckRecord, sizeof (CheckRecord), 1,
			   
Dataset->MarkupFile))
		
return (1);
	      
if (fseek 
		   (Dataset->MarkupFile, -sizeof (CheckRecord), SEEK_CUR))
		
return (1);
	      
NumRecs++;
	    
}
	  
while (CheckRecord.Offset >= Offset);
	  
if (fseek (Dataset->MarkupFile, sizeof (CheckRecord), SEEK_CUR))
	    
return (1);
	  
NumRecs--;
	  
CurrentPos = ftell (Dataset->MarkupFile);
	  
TempRecords = 
	    (MarkupRecord *) calloc (NumRecs, sizeof (MarkupRecord));
	  
if (TempRecords == NULL)
	    
return (2);
	  
if (NumRecs !=
	       
fread (TempRecords, sizeof (MarkupRecord), NumRecs,
		       
Dataset->MarkupFile))
	    
return (1);
	  
if (fseek (Dataset->MarkupFile, CurrentPos, SEEK_SET))
	    
return (1);
	  
if (1 !=
	       
fwrite (&Record, sizeof (MarkupRecord), 1,
			Dataset->MarkupFile))
	    
return (1);
	  
if (NumRecs !=
	       
fwrite (TempRecords, sizeof (MarkupRecord), NumRecs,
			
Dataset->MarkupFile))
	    
return (1);
	
}
      
      else
	
	{
	  
if (1 != fwrite (&Record, sizeof (Record), 1, Dataset->MarkupFile))
	    
return (1);
	  
LastOffset = Offset;
	
}
    
}
  
return (0);

}



//-----------------------------------------------------------------------
// Normalizes a potential heading string in-place.  Returns the length of
// the normalized string.

static int 
NormalizePotentialHeader (char *s) 
{
  
char *ss;
  
int j;
  
    // Normalize the line by removing leading spaces.
    for (ss = s; isspace (*ss); ss++);
  
if (ss != s)
    
strcpy (s, ss);
  
    // Also, remove any trailing spaces or punctuation.
    j = strlen (s);
  
if (j > 0)
    
    {
      
for (ss = &s[j - 1]; ss >= s; ss--)
	
	{
	  
if (isspace (*ss) || ispunct (*ss))
	    
*ss = 0;
	  
	  else
	    
break;
	
}
    
}
  
    // Get rid of multiple spaces.
    for (ss = s; *ss; ss++)
    
if (isspace (*ss))
      
      {
	
*ss = ' ';
	
if (isspace (ss[1]))
	  
	  {
	    
strcpy (ss, ss + 1);
	    
ss--;
	  
}
      
}
  
    // Oh, and let's make it upper-case.
    for (ss = s; *ss; ss++)
    
*ss = toupper (*ss);
  
j = strlen (s);
  
return (j);

}



//---------------------------------------------------------------------------
// Checks a line to see if it's a match against the buffered prefatory lines.

static int 
MatchesPrefatoryLine (AnalysisDataset * Dataset, char *Line) 
{
  
char s[256];
  
int i;
  
if (NULL != Line)
    
    {
      
strcpy (s, Line);
      
NormalizePotentialHeader (s);
      
for (i = 0; i < Dataset->NumNonTrivialPrefatoryLines; i++)
	
if (0 ==
	     
strncmp (s, Dataset->PrefatoryLines[i],
		       
Dataset->PrefatoryLineSizes[i]) 
 &&strlen (s) <
	     2 * Dataset->PrefatoryLineSizes[i])
	  
return (1);
    
}
  
return (0);

}



//------------------------------------------------------------------------
// This function is used to recognize headings.  Returns 0 if it thinks the
// line is not a header.  The array BufferedLines[BUFFERED_LINES] should have
// been pre-loaded.  BufferedLines[PRE_OR_POST_LINES] is the current line, 
// while the other array elements are earlier and later lines from the file.

static int 
IsHeader (LineRecord * BufferedLines, int *LastWasHeader1, int LineNum,
	  
AnalysisDataset * Dataset, char *Line) 
{
  
LineRecord * LineInfo;
  
int InHeader1;
  
int BlankBefore;
  
InHeader1 = 0;
  
LineInfo = BufferedLines + PRE_OR_POST_LINES;
  
BlankBefore = (LineInfo[-1].Empty || LineInfo[0].FormFeedPrior 
		  ||LineInfo[-1].FormFeedAfter);
  
if (BlankBefore && LineInfo[1].Empty && LineInfo[0].AllCaps 
       &&LineInfo[0].OnlyRoman)
    
InHeader1 = 1;
  
if (BlankBefore && LineInfo[1].Empty 
       &&LineInfo[0].AllCaps && LineInfo[0].BeginsChapter)
    
InHeader1 = 1;
  
  else if (*LastWasHeader1)
    
*LastWasHeader1 = 0;
  
  else if (LineNum == 0 && LineInfo[0].Scanned)
    
InHeader1 = 0;
  
  else if (BlankBefore && LineInfo[1].Empty && LineInfo[0].Contents)
    
InHeader1 = 1;
  
  else if (BlankBefore && LineInfo[1].Empty 
	   &&LineInfo[0].AllCaps && !LineInfo[0].EndPeriod)
    
InHeader1 = 1;
  
  else if (BlankBefore && LineInfo[-2].Empty && LineInfo[1].Empty)
    
    {
      

if (LineInfo[-3].Empty && !LineInfo[0].EndPeriod)
	
InHeader1 = 1;
      
if (LineInfo[2].Empty && !LineInfo[0].EndPeriod)
	
InHeader1 = 1;
      
if (LineInfo[0].AllCaps)
	
InHeader1 = 1;
      
if (LineInfo[0].BeginsBook || LineInfo[0].BeginsChapter)
	
InHeader1 = 1;
      
if (LineInfo[0].CapFirstChar && LineInfo[0].BeginsRoman 
	   &&!(LineInfo[0].SpecialRoman && LineInfo[0].WhiteAfterRoman))
	
InHeader1 = 1;
      
if (!LineInfo[0].EndPeriod)
	
InHeader1 = 1;
    
}
  
  else if (BlankBefore && LineInfo[-2].Empty && LineInfo[2].Empty)
    
    {
      
if (MatchesPrefatoryLine (Dataset, Line))
	
InHeader1 = 1;
      
if (LineInfo[0].Short && !LineInfo[1].EndPeriod)
	
InHeader1 = 1;
      
if (LineInfo[-3].Empty && !LineInfo[1].EndPeriod)
	
InHeader1 = 1;
      
if (LineInfo[3].Empty && !LineInfo[1].EndPeriod)
	
InHeader1 = 1;
      
if (LineInfo[0].AllCaps && LineInfo[1].AllCaps)
	
InHeader1 = 1;
      
if (LineInfo[0].BeginsBook || LineInfo[0].BeginsChapter)
	
InHeader1 = 1;
      
if (LineInfo[0].CapFirstChar && LineInfo[0].BeginsRoman 
	   &&!(LineInfo[0].SpecialRoman && LineInfo[0].WhiteAfterRoman))
	
InHeader1 = 1;
      
if (!LineInfo[1].EndPeriod)
	
InHeader1 = 1;
    
}
  
if (!LineInfo[0].VerseCap)
    
InHeader1 = 0;
  
if (LineNum >= Dataset->NumLines - PRE_OR_POST_LINES)
    
InHeader1 = 0;
  
return (InHeader1);

}



//---------------------------------------------------------------------------
// Attempt to find the "prefatory" area, which is what I call the area between
// the PG header and the actual text.  The net result of the pass is simply to
// set Dataset->NumPrefatoryLines.  The basic technique is to find, within the 
// first MAX_PREFATORY_LINES the first candidate heading followed by at least
// MIN_TEXT_LINES of non-headings.  Obviously, some additional flourishes may
// be required.  For example, if we find a non-blank line that's a duplicate 
// of a prior line, it is probably the first section.
//
// The reason I'm doing this at all is that if the Gutenberg text contains a 
// "table of contents", we really messes things up in terms of extraneous 
// headings.  So if I can find something labeled "contents"  or "table of 
// contents", I'll include it even if it's outside of these parameters.

static void 
PrefatoryAnalysisPass (AnalysisDataset * Dataset) 
{
  
int i, j, k, TextLines, LastWasHeader1;
  
LineRecord BufferedLines[BUFFERED_LINES], DefaultLine =
  {
  0};
  
char s[256], *ss;
  

Dataset->LowestPrefatoryLine = 0;
  
Dataset->LowestNonPrefatoryLine = 0;
  

    // See if we can find something labeled as the table of contents.
    // Everything in front of that will be included in the prefatory
    // area by default (but won't be buffered).
    s[sizeof (s) - 1] = 0;
  
fseek (Dataset->InputFile, Dataset->TextStart, SEEK_SET);
  
for (i = 0; i < MAX_PREFATORY_LINES; i++)
    
if (NULL != rfgets (s, sizeof (s) - 1, Dataset->InputFile))
      
      {
	
NormalizePotentialHeader (s);
	
if (0 == strcmp (s, "CONTENTS") 
	     ||0 == strcmp (s, "TABLE OF CONTENTS") 
	     ||0 == strncmp (s, "CONTENTS OF", 11) 
	     ||0 == strncmp (s, "TABLE OF CONTENTS OF", 20))
	  
	  {
	    
Dataset->LowestPrefatoryLine = Dataset->LowestNonPrefatoryLine =
	      
i;
	    
break;
	  
}
      
}
  
    else
      
break;
  

    // Load up the BufferedLines array.
    DefaultLine.OutOfRange = 1;
  
DefaultLine.Empty = 1;
  
fseek (Dataset->LineFile, 0, SEEK_SET);
  
for (i = 0; i < PRE_OR_POST_LINES; i++)
    
BufferedLines[i] = DefaultLine;
  
for (; i < BUFFERED_LINES; i++)
    
    {
      
if (1 !=
	   
fread (BufferedLines + i, sizeof (LineRecord), 1,
		   
Dataset->LineFile))
	
BufferedLines[i] = DefaultLine;
    
}
  

    // Now do the actual search.
    TextLines = -MAX_PREFATORY_LINES;
  
for (i = 0; i < MAX_PREFATORY_LINES; i++)
    
    {
      
if (i >= Dataset->LowestPrefatoryLine)
	
	{
	  
LastWasHeader1 = 0;
	  
if (!BufferedLines[PRE_OR_POST_LINES].Contents 
	       &&IsHeader (BufferedLines, &LastWasHeader1, i, Dataset, NULL))
	    
	    {
	      
Dataset->LowestNonPrefatoryLine = i;
	      
TextLines = 0;
	    
}
	  
	  else
	    
TextLines++;
	  
if (TextLines > MIN_TEXT_LINES)
	    
break;
	
}
      
	// Load the next line.    
	for (j = 0; j < BUFFERED_LINES - 1; j++)
	
BufferedLines[j] = BufferedLines[j + 1];
      
if (1 !=
	   
fread (BufferedLines + BUFFERED_LINES - 1, sizeof (LineRecord), 1,
		   
Dataset->LineFile))
	
BufferedLines[i] = DefaultLine;
    
}
  

if (TextLines <= MIN_TEXT_LINES)
    
Dataset->LowestNonPrefatoryLine = Dataset->LowestPrefatoryLine;
  
fseek (Dataset->LineFile, 0, SEEK_SET);
  

    // Now that the prefatory area has been located, buffer the actual lines
    // from it.  But don't allow duplicates.  If duplicates are found, 
    // then shorten up the area.
    fseek (Dataset->InputFile, Dataset->TextStart, SEEK_SET);
  
s[sizeof (s) - 1] = 0;
  
Dataset->NumNonTrivialPrefatoryLines = 0;
  
for (i = 0; i < Dataset->LowestNonPrefatoryLine; i++)
    
    {
      
rfgets (s, sizeof (s) - 1, Dataset->InputFile);
      
if (i < Dataset->LowestPrefatoryLine)
	
continue;
      
j = NormalizePotentialHeader (s);
      
if (j > 5)
	
	{
	  
	    // Here we have an a non-trivial line.  Let's check to see if we've 
	    // already buffered it.
	    for (k = 0; k < Dataset->NumNonTrivialPrefatoryLines; k++)
	    
if (0 == strcmp (s, Dataset->PrefatoryLines[k]))
	      
	      {
		
		  // It's a duplicate!
		  Dataset->LowestNonPrefatoryLine = i;
		
goto DoneBuffering;
	      
}
	  
	    // Haven't buffered it yet, so it's okay to continue.
	    ss = (char *) calloc (1, j + 1);
	  
if (ss != NULL)
	    
	    {
	      
Dataset->PrefatoryLines[Dataset->NumNonTrivialPrefatoryLines] =
		
ss;
	      
strcpy (ss, s);
	      
Dataset->PrefatoryLineSizes[Dataset->
					   
NumNonTrivialPrefatoryLines] = j;
	      
(Dataset->NumNonTrivialPrefatoryLines)++;
	    
}
	
}
    
}

DoneBuffering:
fseek (Dataset->InputFile, Dataset->TextStart, SEEK_SET);


}



//----------------------------------------------------------------------
// After a certain number of these headings match the ones found in the
// prefatory area, we no longer allow any new headings that DON'T match
// unless they are pretty convincing in some other way.

static int 
NowMatching (AnalysisDataset * Dataset, char *s, LineRecord * Line) 
{
  
int InHeader1 = 1, j;
  
j = MatchesPrefatoryLine (Dataset, s);
  
if (Dataset->NumPrefatoryMatched > PREFATORY_MATCH_CEILING)
    
    {
      
if (!j && !MatchSpecialWords (Dataset, s, Line) && !Line->AllCaps 
	   &&!Line->ReallyShort)
	
InHeader1 = 0;
    
}
  
  else
    
    {
      
if (j)
	
(Dataset->NumPrefatoryMatched)++;
    
}
  
return (InHeader1);

}



//--------------------------------------------------------------------
// Detect a separating line, such as a line of all '*' or '-'.

static int 
IsSeparatorLine (char *s, LineRecord * NextLine) 
{
  
char *ss;
  
if (NextLine->Empty)
    
    {
      
	// First, check for all stars.
	for (ss = s; *ss; ss++)
	
if (!isspace (*ss) && *ss != '*')
	  
break;
      
if (*ss == 0)
	
return (1);
      
	// Now, check for all dashes.
	for (ss = s; *ss; ss++)
	
if (!isspace (*ss) && *ss != '-')
	  
break;
      
if (*ss == 0)
	
return (1);
      
	// ... and underlines.
	for (ss = s; *ss; ss++)
	
if (!isspace (*ss) && *ss != '_')
	  
break;
      
if (*ss == 0)
	
return (1);
      
	// .. and dots.  .
	for (ss = s; *ss; ss++)
	
if (!isspace (*ss) && *ss != '.')
	  
break;
      
if (*ss == 0)
	
return (1);
    
}
  
return (0);

}



//---------------------------------------------------------------------
// Tests if a character is appropriate for end-of-phrase punctuation.

int 
IsEndPunct (char c) 
{
  
return (c == '.' || c == '?' || c == '!' || c == ':' || c == ';'
	   || c == ')' 
 ||c == ',' || c == '}' || c == ']');

}


int 
IsEndSpace (char c) 
{
  
return (!c || isspace (c));

}



//---------------------------------------------------------------------
// The method of analysis uses several passes.  The first pass collects 
// data about all the lines of the file.  The last pass advances through 
// the file a line at a time, keeping in memory a window of the 
// pre-collected line data in which several lines prior to the curent line 
// and several lines after the current line are available for quick access.  
// This might allow, for example, detecting a line preceded and followed 
// by a couple of blank lines as being a header.

int 
MarkBody (AnalysisDataset * Dataset) 
{
  
#define MAX_BACKTRACKS 16
  int InParagraph, InBlockquote, InHeader1, InSubtitle, LastWasHeader1,
    
ErrorCode, InTable, Versifying, LastVersifying;
  
int BracketItalicsCount = 0;
  
char s[256], *ss, *sss, *siv, *sv;
  
char Normalized[MAXWORDLENGTH], Full[MAXWORDLENGTH],
    
TestNormalized[MAXWORDLENGTH], TestFull[MAXWORDLENGTH];
  
int ReturnValue = 1;
  
LineRecord BufferedLines[BUFFERED_LINES], DefaultLine =
  {
  0}
  , *LineInfo;
  
int i, j, k, n;
  
int CouldBeName;
  
int Italicizing = 0, Centering = 0, InWord = 0;
  
int WordItalicizing = 0;	// NOTE:  ALWAYS 0 now.
  int TripSquote = -1;
  
int Foreignosity = 0, LastForeignosity = 0, InPreface;
  
int SentenceStart = 1;
  
enum MarkupType ParagraphType = MarkBeginJustifiedParagraph, Type;
  
struct 
  {
    
long Start;		// Starting offset in etext.
    long End;			// Ending offset in etext.
  }
   
ForeignBacktracks[MAX_BACKTRACKS];
  
int NumBacktracks = 0;
  

for (j = 0; j < NUM_AUTO_ITALICIZED; j++)
    
LengthItalicized[j] = strlen (Italicized[j]);
  

Dataset->LineFile = tmpfile ();
  
if (Dataset->LineFile == NULL)
    
    {
      
fprintf (stderr, "Cannot create first-pass temporary file.\n");
      
ReturnValue = 2;
      
goto Done;
    
}
  

    //---------------------------------------------------------------------
    // Run the first analysis pass.
    if (0 != (ErrorCode = LineAnalysisPass (Dataset)))
    
    {
      
fprintf (stderr, "Line-analysis failed (code = 0x%X).\n", ErrorCode);
      
ReturnValue = 3;
      
goto Done;
    
}
  

    //----------------------------------------------------------------------
    // Another analysis pass that attempts to determine where the title-page &
    // contents are.  In other words, the area between the PG header and the
    // actual text. 
    PrefatoryAnalysisPass (Dataset);
  
Dataset->NumPrefatoryMatched = 0;
  

    //----------------------------------------------------------------------
    // Run the Final analysis pass.
    
    // Set up the data structures.
    s[sizeof (s) - 1] = '\0';
  
fseek (Dataset->LineFile, 0, SEEK_SET);
  
fseek (Dataset->InputFile, Dataset->TextStart, SEEK_SET);
  
DefaultLine.OutOfRange = 1;
  
DefaultLine.Empty = 1;
  
for (i = 0; i < PRE_OR_POST_LINES; i++)
    
BufferedLines[i] = DefaultLine;
  
for (; i < BUFFERED_LINES; i++)
    
    {
      
if (1 !=
	   
fread (BufferedLines + i, sizeof (LineRecord), 1,
		   
Dataset->LineFile))
	
BufferedLines[i] = DefaultLine;
    
}
  
InParagraph = InBlockquote = InHeader1 = LastWasHeader1 = InSubtitle =
    
InTable = Versifying = LastVersifying = 0;
  
LineInfo = &BufferedLines[PRE_OR_POST_LINES];
  

    // Okay, here's where we actually loop through the lines of the input file.
    for (i = 0; i < Dataset->NumLines; i++)
    
    {
      
InPreface = (i < Dataset->LowestNonPrefatoryLine);
      
	// Fetch the current line of the input file.  The window of buffered
	// LineRecord data is already set up properly.
	if (NULL == rfgets (s, sizeof (s) - 1, Dataset->InputFile))
	
	{
	  
fprintf (stderr, "Premature end of input file.\n");
	  
ReturnValue = 4;
	  
goto Done;
	
}
      

	// Analyze.  In case it isn't obvious, LineInfo[0] contains the
	// LineRecord for the current line.  LineInfo[-1], ..., 
	// LineInfo[-PRE_OR_POST_LINES] are for the prior lines.  
	// LineInfo[1], ..., LineInfo[PRE_OR_POST_LINES]
	// are for the succeeding lines.
	
	// Locate PG file-ender.  We define this as any line within
	// 500 characters of the file-end that contains the words
	// "end", and "project gutenberg" or "PG".
	if (ftell (Dataset->InputFile) > Dataset->FileEnderRegion)
	
	{
	  
int End = 0, PG = 0;
	  
for (ss = s; *ss; ss++)
	    
	    {
	      
if (!strncasecmp (ss, "end ", 4))
		
		{
		  
End = 1;
		  
if (PG)
		    
break;
		
}
	      
	      else if (!strncasecmp (ss, " PG ", 4) 
		       ||!strncasecmp (ss, "project gutenberg", 17))
		
		{
		  
PG = 1;
		  
if (End)
		    
break;
		
}
	    
}
	  
if (End && PG)
	    
	    {
	      
if (AddMarkup 
		   (Dataset, LineInfo[0].Offset, MarkGutenbergEnder, 0))
		
goto DiskError;
	    
}
	
}
      

	// *** Do stuff pertaining to the line properties, such as 
	// paragraph begin/end, headers, and verse ***
	
if (InBlockquote)
	
	{
	  
if (LineInfo[0].Empty || !LineInfo[0].BeginsWhite)
	    
	    {
	      
InBlockquote = 0;
	      
if (AddMarkup 
		   (Dataset, LineInfo[0].Offset, MarkEndBlockquote, 0))
		
goto DiskError;
	    
}
	
}
      
if (InSubtitle)
	
	{
	  
if (LineInfo[0].Empty)
	    
	    {
	      
InSubtitle = 0;
	      
if (AddMarkup
		   (Dataset, LineInfo[0].Offset, MarkEndSubtitle, 0))
		
goto DiskError;
	    
}
	
}
      
if (InTable)
	
	{
	  
if (LineInfo[0].Empty)
	    
	    {
	      
InTable = 0;
	      
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkEndTable, 0))
		
goto DiskError;
	    
}
	
}
      
if (InParagraph)
	
	{
	  
if (LineInfo[0].Empty /* || LineInfo[0].BeginsWhite */ )
	    
	    {
	      
		// End the paragraph.  For safety's sake, assume that 
		// italicizing doesn't cross paragraphs, if it's of the 
		// delimited kind.  An exception for the delimited kind 
		// is if <I> was used.  We assume in that case
		// that the user had enough on the ball to get it right.
		if (Italicizing != 0 && Italicizing != 'A' 
		    &&Italicizing != 'H')
		
		{
		  
Italicizing = 0;
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset, MarkEndItalics, 0))
		    
goto DiskError;
		
}
	      
InParagraph = 0;
	      
if (AddMarkup 
		   (Dataset, LineInfo[0].Offset, MarkEndParagraph, 0))
		
goto DiskError;
	    
}
	  
	  else if (Versifying)
	    
	    {
	      
if (LineInfo[0].VerseCap || !LineInfo[0].BeginsWhite 
		   ||(!LineInfo[0].VerseCap && LineInfo[0].Leading > 1))
		
		{
		  
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkBreak, 0))
		    
goto DiskError;
		  
for (j = 0; j < LineInfo[0].Leading; j++)
		    
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkNbsp, 0))
		      
goto DiskError;
		
}
	    
}
	  
	  else if (LineInfo[-1].Leading > 1 || LineInfo[0].Leading > 1 
		   ||(LineInfo[-1].Leading == 1 && LineInfo[0].Leading == 1) 
		   ||(LineInfo[0].Leading == 1 && !LineInfo[-1].Empty 
		      &&LineInfo[-2].Leading == 1) 
		   ||(LineInfo[0].Leading == 1 && !LineInfo[1].Empty 
		      &&LineInfo[2].Leading == 1) 
		   ||(LineInfo[-1].Leading == 1 && !LineInfo[-2].Empty 
		      &&LineInfo[-3].Leading == 1) 
		   ||(LineInfo[-1].Leading == 1 && !LineInfo[0].Empty 
		      &&LineInfo[1].Leading == 1))
	    
	    {
	      
		// The point of this to to detect intentionally indented lines
		// in the middle of a paragraph.  But we assume that a single 
		// line with a single space at the front -- isolated, without
		// other such lines in the vicinity -- is a simple mistake.
		// However, some verse has every OTHER line indented, so we 
		// try to detect that.
		if (AddMarkup (Dataset, LineInfo[0].Offset, MarkBreak, 0))
		
goto DiskError;
	      
if (!Centering)
		
		{
		  
for (j = 0; j < LineInfo[0].Leading; j++)
		    
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkNbsp, 0))
		      
goto DiskError;
		
}
	    
}
	  
	  else if (i < Dataset->LowestNonPrefatoryLine)
	    
	    {
	      
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkBreak, 0))
		
goto DiskError;
	    
}
	  
	  else if (!LineInfo[-1].Empty && LineInfo[-1].Short 
		   &&(LineInfo[-2].Empty || LineInfo[-2].Short) 
		   &&(LineInfo[0].Short 
		      ||(LineInfo[1].Empty || LineInfo[1].Short)) 
		   &&ParagraphType != MarkBeginJustifiedParagraph)
	    
	    {
	      
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkBreak, 0))
		
goto DiskError;
	    
}
	  
	  else if (LineInfo[0].VerseCap)
	    
	    {
	      
		// Hopefully, what with the "Versifying" variable (which 
		// didn't exist previously), this code shouldn't be needed.  
		// But heck, it was here!  What we do here is to try and 
		// detect a line of verse which is nevetheless longer than 
		// the arbitrary limit of LineInfo[0].Short.  (This is 
		// actually a pretty rare condition, but heck!)
		// We do this by noting that lines of verse are generally 
		// initially capitalized, so we look for a whole block of 
		// lines that meet this criterion. 
		k = 1;
	      
for (j = -1; j >= -PRE_OR_POST_LINES; j--)
		
if (LineInfo[j].VerseCap)
		  
k++;
	      
		else if (!LineInfo[j].Empty)
		  
break;
	      
if (j >= -PRE_OR_POST_LINES && !LineInfo[j].Empty)
		
k = -1000;
	      
for (j = 1; j <= PRE_OR_POST_LINES; j++)
		
if (LineInfo[j].VerseCap)
		  
k++;
	      
		else if (!LineInfo[j].Empty)
		  
break;
	      
if (j <= PRE_OR_POST_LINES && !LineInfo[j].Empty)
		
k = -1000;
	      
if (k >= 4)
		
		{
		  
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkBreak, 0))
		    
goto DiskError;
		
}
	    
}
	
}
      
if (InHeader1 && LineInfo[-1].Empty)
	
	{
	  
InHeader1 = 0;
	  
LastWasHeader1 = 1;
	  
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkEndHeader1, 0))
	    
goto DiskError;
	
}
      
if (!InParagraph && !InBlockquote && !InHeader1 && !InSubtitle 
	   &&!InTable && !LineInfo[0].Empty)
	
	{
	  
	    // LastVersifying is a variable that grows slowly while versifying, 
	    // but decays quickly when not.  It allows a sort of persistence of
	    // detecting verse when intervening stuff is thrown in (like
	    // footnotes, dividers "****", and so on).
	    if (Versifying)
	    
LastVersifying++;
	  
	  else if (LastWasHeader1)
	    
LastVersifying--;
	  
	  else
	    
LastVersifying /= 2;
	  
Versifying = 0;
	  
if (LineInfo[0].BeginsWhite)
	    
	    {
	      
		// If the line began with whitespace, we never count it as a 
		// potential heading, unless it matches certain string patterns.
		j = LastWasHeader1;
	      
if (i > Dataset->LowestNonPrefatoryLine)
		
if (MatchesPrefatoryLine (Dataset, s) 
		     ||MatchSpecialWords (Dataset, s, &LineInfo[0]))
		  
		  {
		    
if (IsHeader 
			 (BufferedLines, &LastWasHeader1, i, Dataset, s))
		      
		      {
			
if (NowMatching (Dataset, s, &LineInfo[0]))
			  
goto StartHeader1;
		      
}
		    
		    else if (j)
		      
		      {
			
j = 0;
			
if (MatchesPrefatoryLine (Dataset, s) 
			     ||IsHeader (BufferedLines, &j, i, Dataset, s))
			  
			  {
			  
BeginSubtitle:
InSubtitle =
			      1;
			    
for (ss = s; isspace (*ss); ss++);
			    
if (AddMarkup 
				 (Dataset, LineInfo[0].Offset + ss - s,
				  
MarkBeginSubtitle, 0))
			      
goto DiskError;
			    
goto ParagraphBegun;
			  
}
		      
}
		  
}
	      
		// Here we try to detect block quotes.  We do this by 
		// determining that all of the lines are indented the 
		// same amount, and that they don't begin all capitalized.
		k = LineInfo[0].Leading;
	      
if (LineInfo[0].BeginsQuote && LineInfo[1].Leading == k + 1)
		
k++;
	      
n = LineInfo[0].VerseCap;
	      
for (j = 1; j <= PRE_OR_POST_LINES; j++)
		
		{
		  
if (LineInfo[j].Empty)
		    
break;
		  
if (LineInfo[j].Leading != k)
		    
break;
		  
n += LineInfo[j].VerseCap;
		
}
	      
if ((j > PRE_OR_POST_LINES || LineInfo[j].Empty) && j > 3 
		   &&n != j)
		
		{
		  
InBlockquote = 1;
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset, MarkBlockquote, 0))
		    
goto DiskError;
		  
goto ParagraphBegun;
		
}
	      
		// Well, not recognized as a block quote. 
		Type = MarkBeginRaggedParagraph;
	    
StartParagraph:
if (!Versifying)
		
		{
		  
		    // Is it the start of a table?  We'd LIKE to detect 
		    // these simply as paragraphs all of whose lines are 
		    // "PossibleTable" data. However, it's possible that the 
		    // column-headers might occupy multiple rows, and if one
		    // of the column-headers is one or more lines longer than 
		    // the others, it can't be recognized as tabular data.  
		    // So we have to count these specially.  For tables 
		    // containing blank lines -- well, we're just out of 
		    // luck there.
		    for (j = 0, k = -1; j <= PRE_OR_POST_LINES; j++)
		    
if (LineInfo[j].Empty)
		      
break;
		  
		    else if (!LineInfo[j].PossibleTable)
		      
		      {
			
if (k == j - 1)
			  
k++;
			
			else
			  
break;
		      
}
		  
if (j > k + 2 
		       &&(j > PRE_OR_POST_LINES || LineInfo[j].Empty))
		    
		    {
		      
InTable = 1;
		      
if (AddMarkup 
			   (Dataset, LineInfo[0].Offset, MarkBeginTable, 0))
			
goto DiskError;
		      
goto ParagraphBegun;
		    
}
		
}
	      
		// No, just a regular paragraph.  
		InParagraph = 1;
	      
		// Try to detect centered text.  There are two conditions we
		// apply.  First, at least one line in the current contiguous
		// group is ragged.  Second, we count the number of ragged lines
		// in the vicinity, and determine that the count is greater than
		// a certain number.  An exception to the latter is if every
		// line in the contiguous group is ragged.  Lastly, if the 
		// paragraph contains a single line, the RaggedStart criterion 
		// is too limiting, so in that case we look for any white space.  
		// All of this applies only if the sole line that's indented is 
		// NOT the first line of the paragraph.  Oh, and don't center 
		// verse:
		if (Versifying)
		
		{
		  
Type = MarkBeginRaggedParagraph;
		  
goto TypeKnown;
		
}
	      
		// Here we detect a simple indented paragraph -- we think!
		if (LineInfo[0].BeginsWhite && !LineInfo[1].Empty)
		
		{
		  
for (j = 1; j <= PRE_OR_POST_LINES; j++)
		    
if (LineInfo[j].Empty || LineInfo[j].BeginsWhite)
		      
break;
		  
if (j > PRE_OR_POST_LINES || LineInfo[j].Empty)
		    
		    {
		      
			// Okay, it's just an indented paragraph.  
			// Go to the regular
			// analysis.
			LineInfo[0].BeginsWhite = 0;
		      
LineInfo[0].Leading = 0;
		      
goto RegularParagraph;
		    
}
		
}
	      
		// Here we look for a line whose purpose is simply to 
		// separate blocks of text.
		if (IsSeparatorLine (s, &LineInfo[1]))
		
		{
		  
Type = MarkBeginCenteredParagraph;
		  
goto TypeKnown;
		
}
	      
		// Here we look for a single contiguous group of lines 
		// that is ragged.    
		for (k = j = 0; j < PRE_OR_POST_LINES; j++)
		
		{
		  
if (LineInfo[j].Empty)
		    
break;
		  
if (LineInfo[j].RaggedStart 
		       ||(j == 0 && LineInfo[j].BeginsWhite))
		    
k++;
		
}
	      
if (k == j)	// Here, EVERY line was ragged.
		{
		  
		    // One problem is that the set of contiguous lines 
		    // might have been really short -- maybe just one line.  
		    // Maybe it was just an indented paragraph, if it's only 
		    // one line.  In that case, we check the vicinity for 
		    // additional ragged lines, unless the indentation is 
		    // really big!
		    if (j == 1)
		    
		    {
		      
if (LineInfo[0].Leading > 10)
			
Type = MarkBeginCenteredParagraph;
		      
		      else
			
goto MoreRaggedChecking;
		    
}
		  
		  else
		    
Type = MarkBeginCenteredParagraph;
		
}
	      
	      else if (k > 0)	// Here, just SOME lines were ragged.
		{
		
MoreRaggedChecking:
		    // We check additional lines from the vicinity, but 
		    // possibly outside the contiguous group, to see if 
		    // they're ragged too.
		    for (k = 0, j = -PRE_OR_POST_LINES;
			 j <= PRE_OR_POST_LINES; 
j++)
		    
		    {
		      
if (LineInfo[j].RaggedStart 
			   ||(j == 0 && LineInfo[j].BeginsWhite))
			
k++;
		    
}
		  
if (k >= 2)
		    
Type = MarkBeginCenteredParagraph;
		
}
	    
TypeKnown:
Centering =
		(Type == MarkBeginCenteredParagraph);
	      
ParagraphType = Type;
	      
if (AddMarkup (Dataset, LineInfo[0].Offset, ParagraphType, 0))
		
goto DiskError;
	      
if (!Centering)
		
		{
		  
for (ss = s; isspace (*ss); ss++)
		    
if (AddMarkup (Dataset, LineInfo[0].Offset, MarkNbsp, 0))
		      
goto DiskError;
		
}
	    
ParagraphBegun:;
	    
}
	  
	  else
	    
	    {
	      
if (i > 0 && i == Dataset->LowestNonPrefatoryLine)
		
InHeader1 = 1;
	      
	      else if (i < Dataset->LowestNonPrefatoryLine)
		
InHeader1 = 0;
	      
	      else
		
		{
		  
j = LastWasHeader1;
		  
InHeader1 =
		    
IsHeader (BufferedLines, &LastWasHeader1, i, Dataset, s);
		  
if (j && !InHeader1)
		    
		    {
		      
j = 0;
		      
if (MatchesPrefatoryLine (Dataset, s) 
			   ||IsHeader (BufferedLines, &j, i, Dataset, s))
			
goto BeginSubtitle;
		    
}
		
}
	      
if (InHeader1)
		
InHeader1 = NowMatching (Dataset, s, &LineInfo[0]);
	      
if (InHeader1)
		
		{
		
StartHeader1:
InHeader1 = 1;
		  
		    // NOTE THAT THE FOLLOWING MARKUP HAS THE POTENTIAL TO BE 
		    // OUT OF ORDER IF SOME TYPES OF CODE CHANGES OCCUR LATER!
		    for (ss = s; isspace (*ss); ss++);
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset + ss - s, MarkHeader1, 0))
		    
goto DiskError;
		
}
	      
	      else
		
		{
		
RegularParagraph:
		    // Here's where the analysis starts for a plain-Jane 
		    // unindented Gutenberg paragraph.
		    
		    // We know that we want to start a paragraph, but we don't 
		    // know if it should be justified or ragged.  
		    
		    // Here we look for a line whose purpose is simply to 
		    // separate blocks of text. 
		    if (IsSeparatorLine (s, &LineInfo[1]))
		    
		    {
		      
Type = MarkBeginCenteredParagraph;
		      
goto TypeKnown;
		    
}
		  

		    // Let's check to see if the lines seem to be shorter 
		    // than expected.  Note that we count some conditions 
		    // twice as much as those that are simply short.
		    // We do it intentionally.
		    if (i < Dataset->LowestNonPrefatoryLine)
		    
Type = MarkBeginRaggedParagraph;
		  
		  else
		    
		    {
		      
			// Here we look for long chunks of obvious verse, 
			// which we detect as complete paragraphs (to the 
			// extent we can determine such a thing), 
			// every line of which begins with caps.  However, 
			// we reject short paragraphs with lines beginning
			// with quotes from the count.
			for (j = k = 0; j < PRE_OR_POST_LINES; j++)
			
if (LineInfo[j].Empty)
			  
break;
		      
			else if (LineInfo[j].VerseCap 
#ifdef DOESNT_SEEM_TO_DO_WHAT_I_EXPECTED
				 &&!(LineInfo[j].BeginsQuote 
				     &&((j < PRE_OR_POST_LINES - 1 
					 &&LineInfo[j + 1].Empty) 
					||(j > 0 
					   &&LineInfo[j - 1].Empty))) 
#endif	/* 
 */
			  )
			  
k++;
		      
if (k == j)
			
			{
			  
if (j > 5)
			    
Versifying = 1;
			  
			  else if (j > 3 && LastVersifying)
			    
Versifying = 1;
			  
			  else if (j > 2 && LastVersifying > 2)
			    
Versifying = 1;
			
}
		      

			// Sort of a catch-all heuristic thing to detect verse
			// that wasn't so obvious as the above.
			if (!Versifying)
			
for (j = 0, k = (LastVersifying > 0);
			      
j < PRE_OR_POST_LINES; j++)
			  
			  {
			    
if (LineInfo[j].Empty)
			      
break;
			    
if (LineInfo[j].BeginsWhite 
				 ||(LineInfo[j].ReallyShort 
				    &&!LineInfo[j + 1].Empty 
				    &&LineInfo[j].VerseCap))
			      
k += 2;
			    
			    else if (LineInfo[j].Short 
				     &&!LineInfo[j + 1].Empty 
				     &&LineInfo[j].VerseCap)
			      
k++;
			    
if (!LineInfo[j].BeginsWhite 
				 &&!LineInfo[j].VerseCap)
			      
k--;
			  
}
		      
			// The above rules are fine, but OCCASIONALLY
			// catch a line of dialog, because when there's 
			// dialog there are usually a lot of 1- or 2-line
			// paragraphs, possibly short, all beginning with
			// caps.  The line below just reduces that tendency
			// a little.        
			if (j <= 2 && !LineInfo[0].BeginsWhite && 
			    (LineInfo[0].BeginsQuote 
			     ||LineInfo[-2].BeginsQuote 
			     ||LineInfo[2].BeginsQuote 
			     ||(LineInfo[2].Empty
				&& LineInfo[3].BeginsQuote) 
 ||LineInfo[1].
			     Empty))
			
k--;
		      
if (Versifying || (k >= 2 && k >= (j - 1) / 2) 
			   ||(j == 1 && LineInfo[0].Short) || k == j 
			   ||k == (j - 1))
			
			{
			  
Versifying = 1;
			  
Type = MarkBeginRaggedParagraph;
			
}
		      
		      else
			
Type = MarkBeginJustifiedParagraph;
		    
}
		  
goto StartParagraph;
		
}
	    
}
	
}
      

	// *** Now do stuff pertaining to individual characters, such as 
	// adding extra space between sentences, eliminating hyphens, 
	// replacing "--" with long hyphens, etc. ***
	
if (!Italicizing)
	
WordItalicizing = 0;
      
for (ss = s; *ss; ss++)
	
	{
	  
	    // SentenceStart is a helpful flag for determining if 
	    // a word may be at the start of a sentence.  It's used
	    // ONLY for the all-caps italicizaton style.
	    SentenceStart |= *ss == '.' || *ss == '?' || *ss == '!' 
	    ||*ss == ':' || *ss == '\"';
	  
if (SentenceStart)
	    
NumBacktracks = Foreignosity = 0;
	  
	    // Pick off words.
	    if (InWord)
	    
	    {
	      
if (IsWordChar (*ss))
		
		{
		  
if (WordItalicizing && *ss != '\'')
		    
		    {
		      
if (AddMarkup 
			   (Dataset, LineInfo[0].Offset + ss - s, MarkTolower,
			    
0))
			
goto DiskError;
		    
}
		
}
	      
	      else
		
		{
		  
InWord = 0;
		  
if (WordItalicizing && Italicizing)
		    
		    {
		      
WordItalicizing = Italicizing = 0;
		      
if (AddMarkup 
			   (Dataset, LineInfo[0].Offset + ss - s,
			    
MarkEndItalics, 0))
			
goto DiskError;
		    
}
		
}
	    
}
	  
	    // We don't allow any WORD-based processing in the Prefatory area,
	    // because it's pointless.  
	    if (!InWord && IsWordChar (*ss) && *ss != '\'' && !InPreface)
	    
	    {
	      
char *sss;
	      
int i;
	      
int Matched;
	      
		// Locate the end of the word.
		for (sss = ss + 1;; sss++)
		
		{
		  
if (!*sss)
		    
break;
		  
if (!IsWordChar (*sss))
		    
break;
		  
if (*sss == '\'')
		    
		    {
		      
i = IsWordChar (sss[1]);
		      
if (i == WORD_NOT || 0 != (i & WORD_PUNCT))
			
break;
		    
}
		
}
	      
		// Process the word.
		if (sss == ss + 1)
		
NumBacktracks = Foreignosity = 0;
	      
if (sss > ss + 1 && sss < ss + MAXWORDLENGTH)
		
		{
		  
InWord = 1;
		  
strncpy (Full, ss, sss - ss);
		  
Full[sss - ss] = 0;
		  
if (DiacriticalNormalize (Full, Normalized, MAXWORDLENGTH))
		    
		    {
		      
			// Because various of the markups which the analyses
			// below want to trigger conflict with each other,
			// we can't just start the markups when the analyses
			// identify the need for them.  Instead, we have to
			// set some state variables instead, and then after
			// all of the analyses are done resolve the conflicts.
			// Only then can we apply markup.
		      int WantToItalicize = 0;
		      
int WantFirstCharLower = 0;
		      
int WantRestOfWordLower = 0;
		      
int Want8Bit = 0;
		      
			// Process ALL-CAPS words against the wordlist created
			// from the etext, for the purpose of converting to 
			// lower case and italicizing.  This is only interesting
			// if we're not already italicizing. 
			if (!Italicizing && IsStrupr (Full) 
			    &&!InHeader1 && !InSubtitle)
			
			{
			  
k = SentenceStart;
			  
strcpy (TestFull, Full);
			  
strcpy (TestNormalized, Normalized);
			  
DiacriticalStrlwr (&TestFull[SentenceStart]);
			  
DiacriticalStrlwr (&TestNormalized[SentenceStart]);
			  
SearchWordlist (Dataset->Words, TestNormalized,
					   
TestFull, &Matched);
			  
if (!Matched && !SentenceStart)
			    
			    {
			      
TestFull[0] = toupper (TestFull[0]);
			      
TestNormalized[0] =
				toupper (TestNormalized[0]);
			      
k = 1;
			      
SearchWordlist (Dataset->Words, TestNormalized,
					       
TestFull, &Matched);
			    
}	// !Matched && ...
			  if (Matched)
			    
			    {
			      
				//WordItalicizing = Italicizing = 1;
				//if (AddMarkup
				//(Dataset, LineInfo[0].Offset + ss - s,
				//MarkBeginItalics, 0))
				//goto DiskError;
				WantToItalicize = 1;
			      
WantRestOfWordLower = 1;
			      
if (!k)
				
				{
				  
				    //if (AddMarkup
				    //(Dataset, LineInfo[0].Offset + ss - s,
				    //MarkTolower, 0))
				    //goto DiskError;
				    WantFirstCharLower = 1;
				
}	// !k
			    }	// Matched
			}	// !Italicizing && ...
		      // Here we check the word against the global
		      // wordlists/namelists.  This helps us to improve
		      // ALL-CAPS italicizing, for words not found in the
		      // etext-derived wordlist above.  Also, we can determine
		      // if the word is foreign, or if it's a 7-bit form that
		      // needs to be replaced with an 8-bit form.  In theory,
		      // we could also remove soft-hyphens, but I'll leave
		      // this for the future (if ever).
		      i =
			SearchWordlist (Dataset->Words, Normalized, 
Full,
					&Matched);
		      
if (Matched)	// This can't fail, but still ...
			{
			  
j = Dataset->Words->Words[i].WordlistStatus;
			  
if (Dataset->Words->Words[i].Frequent)
			    
j &= ~SPELL_NONNATIVE;
			  
			    // If we're not already italicizing, we begin 
			    // italicizing for appropriate ALL-CAPS situations.
			    if (!Italicizing && IsStrupr (Full) 
				&&!InHeader1 
				&&!InSubtitle /* !LineInfo[0].AllCaps */ )
			    
			    {
			      
if (0 != (j & SPELL_LOWERCASE)
				   || 
0 != (j & SPELL_CAPITALIZED))
				
				{
				  
k =
				    SentenceStart | (j & SPELL_CAPITALIZED);
				  
				    //WordItalicizing = Italicizing = 1;
				    //if (AddMarkup
				    //(Dataset, LineInfo[0].Offset + ss - s,
				    //MarkBeginItalics, 0))
				    //goto DiskError;
				    WantToItalicize = 1;
				  
WantRestOfWordLower = 1;
				  
if (!k)
				    
				    {
				      
					//if (AddMarkup
					//(Dataset,
					//LineInfo[0].Offset + ss - s,
					//MarkTolower, 0))
					//goto DiskError;
					WantFirstCharLower = 1;
				    
}	// !k
				}	// 0 != ...
			    }	// !Italicizing && ...
			  // Next, do foreign words.  This is seemingly a lot 
			  // easier, because we are not so worried about 
			  // capitalization, but it's actually pretty tricky.
			  // The problem is that many foreign words are also
			  // recognizable as names or as words in the native
			  // language of the etext, and hence we can easily
			  // get into a situation where a foreign phrase shows
			  // up as a mixture of italicized and non-italicized
			  // words (which is very annoying).  Therefore, a 
			  // more complex recognition formula is needed.  What
			  // we do is to compute an index of "foreignosity",
			  // only partly related to specific words, and 
			  // italicize on the basis of this index.  We have
			  // two data at our disposal, for individual words:
			  // the SPELL_FOREIGN flag (which is set if the word
			  // has ONLY been recognized as foreign) and the 
			  // SPELL_NONNATIVE flag (which is set if the word is
			  // found in a foreign dictionary even if it was 
			  // previously found in a native dictionary).
			  // Another difficulty, is that there's a tendency for
			  // the beginnings of foreign phrases to be left off,
			  // if the words correspond to native words.  So once
			  // we reach an appropriate foreignosity threshhold,
			  // we have to work backward to account for them. 
			  LastForeignosity = Foreignosity;
			  
if (SentenceStart)
			    
Foreignosity = 0;	//Foreignosity /= 4;
			  if (0 != (j & SPELL_FOREIGN))
			    
Foreignosity = 256;	//Foreignosity += 256;
			  else if (0 != (j & SPELL_NONNATIVE))
			    
;	//Foreignosity += 16;
			  else
			    
Foreignosity = 0;
			  
if (!Italicizing && !Dataset->NoForeign)
			    
			    {
			      
if (Foreignosity >=
				   
256 /* && 0 != (j & SPELL_NONNATIVE) */ )
				
				{
				  
if (LastForeignosity < 256)
				    
				    {
				      
for (k = 0; k < NumBacktracks; k++)
					
					{
					  
					    // This markup stuff is okay,
					    // because it applies not to 
					    // the current word, but to 
					    // prior words where conflicts
					    // have already been resolved.
					    if (AddMarkup 
						(Dataset,
						 
ForeignBacktracks[k].Start,
						 
MarkBeginItalics, 0))
					    
goto DiskError;
					  
if (AddMarkup 
					       (Dataset,
						
ForeignBacktracks[k].End,
						
MarkEndItalics, 0))
					    
goto DiskError;
					
}
				    

}
				  
NumBacktracks = 0;
				  
				    //WordItalicizing = Italicizing = 1;
				    //if (AddMarkup
				    //(Dataset, LineInfo[0].Offset + ss - s,
				    //MarkBeginItalics, 0))
				    //goto DiskError;
				    WantToItalicize = 1;
				
}	// Foreignosity ...
			      else if (0 != (j & SPELL_NONNATIVE))
				
				{
				  
if (NumBacktracks < MAX_BACKTRACKS)
				    
				    {
				      
ForeignBacktracks[NumBacktracks].
					Start = 
LineInfo[0].Offset + ss - s;
				      
ForeignBacktracks[NumBacktracks].End =
					
ForeignBacktracks[NumBacktracks]. 
					Start + strlen (Full);
				      
NumBacktracks++;
				    
}
				
}	// 0 != ...
			      else
				
NumBacktracks = 0;
			    
}	// !Italicizing.
			  else
			    
NumBacktracks = 0;
			
}	// Matched                            
		      // Now, handle restoration of diacriticals, by replacing
		      // 7-bit ASCII with 8-bit ASCII.   
		      if (0 != (j & SPELL_NORMALIZED) 
			  &&Dataset->Words->Words[i].Match != NULL 
			  &&!Dataset->NoDiacritical)
			
Want8Bit = 1;
		      
			// Now, apply all of the markups implied by the flags we just set.
			k = LineInfo[0].Offset + ss - s;
		      
if (WantToItalicize)
			
			{
			  
if (AddMarkup (Dataset, k, MarkBeginItalics, 0))
			    
goto DiskError;
			
}
		      
if (Want8Bit)
			
sss = Dataset->Words->Words[i].Match;
		      
		      else
			
sss = Dataset->Words->Words[i].Full;
		      
if (WantFirstCharLower || WantRestOfWordLower 
			   ||Want8Bit)
			
			{
			  
for (n = 0; *sss; sss++, n++)
			    
			    {
			      
char c;
			      
c = *sss;
			      
if (!n)
				
				{
				  
if (WantFirstCharLower)
				    
c = DiacriticalTolower (c);
				
}
			      
			      else
				
				{
				  
if (WantRestOfWordLower)
				    
c = DiacriticalTolower (c);
				
}
			      
if (c == '\'')
				
				{
				  
if (AddMarkup (Dataset, k, MarkRsquo, c))
				    
goto DiskError;
				
}
			      
			      else
				
				{
				  
if (AddMarkup 
				       (Dataset, k, MarkInsertChar, c))
				    
goto DiskError;
				
}
			    
}
			  
for (sss = Dataset->Words->Words[i].Full; *sss;
				
sss++, k++)
			    
if (AddMarkup (Dataset, k, MarkRemoveChar, 0))
			      
goto DiskError;
			  
TripSquote = k;
			
}
		      
		      else
			
k += strlen (Full);
		      
if (WantToItalicize)
			
			{
			  
if (AddMarkup (Dataset, k, MarkEndItalics, 0))
			    
goto DiskError;
			
}
		    
}		// DiacriticalNormalize ...
		}		// for ...
	      SentenceStart = 0;
	    
}
	  
	    // "Smart" double-quotes.  
	    if (*ss == '\"' && !IsEndSpace (ss[1]) 
		&&(ss == s || isspace (ss[-1])))
	    
	    {
	      
if (AddMarkup 
		   (Dataset, LineInfo[0].Offset + ss - s, MarkBeginSmartQuote,
		    
0))
		
goto DiskError;
	      
continue;
	    
}
	  
if (*ss == '\"' && ss > s && !isspace (ss[-1]) && 
	       (IsEndSpace (ss[1]) 
		||(IsEndPunct (ss[1]) && IsEndSpace (ss[2]))))
	    
	    {
	      
if (AddMarkup 
		   (Dataset, LineInfo[0].Offset + ss - s, MarkEndSmartQuote,
		    
0))
		
goto DiskError;
	      
continue;
	    
}
	  
	    // "Smart" single-quotes.  This differs a little from 
	    // double-quotes, because we interpret ALL single-quotes as
	    // right-quotes UNLESS they fulfill reasonable criteria for
	    // being a left-hand quote AND an ending single quote appears in
	    // the not-too distant future (defined as the end of the line).
	    if (*ss == '\'')
	    
	    {
	      
int Right = 1;
	      
if (!IsEndSpace (ss[1]) 
		   &&(ss == s || isspace (ss[-1]) || ss[-1] == '\"'))
		
		{
		  
char *sss;
		  
		    // Reasonable requirements for an opening quote are
		    // fulfilled.  Now let's look to see if we can find a
		    // trailing quote prior to the next candidate for opening
		    // quote.
		    for (sss = ss + 1; *sss; sss++)
		    
if (*sss == '\'')
		      
		      {
			
if (!IsEndSpace (sss[1]) 
			     &&(isspace (sss[-1]) || sss[-1] == '\"'))
			  
break;
			
if (!isspace (sss[-1]) && 
			     (IsEndSpace (sss[1]) || sss[1] == '\"' 
			      ||(IsEndPunct (sss[1]) 
				 &&(IsEndSpace (sss[2]) || sss[1] == '\"'))))
			  
			  {
			    
Right = 0;
			    
break;
			  
}
		      
}
		
}
	      
if (Right)
		
		{
		  
k = LineInfo[0].Offset + ss - s;
		  
		    // It may be that this quote was already added as part
		    // of a WORD, and so adding new markup for it would
		    // result in the appearance of two quotes.
		    if (k >= TripSquote)
		    
		    {
		      
if (AddMarkup (Dataset, k, MarkRsquo, 0))
			
goto DiskError;
		    
}
		
}
	      
	      else
		
		{
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset + ss - s, MarkLsquo, 0))
		    
goto DiskError;
		
}
	      
continue;
	    
}
	  
	    // Find multiple dashes.  Replace even-length strings of
	    // dashes with half as many mdashes, and odd-length strings
	    // (>1) with mdashes plus a trailing ndash.  We also convert
	    // " - " to " &mdash; ".
	    if (*ss == '-' && !Dataset->NoMdash)
	    
	    {
	      
if (ss > s && isspace (ss[-1]) && isspace (ss[1]))
		
		{
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset + ss - s, MarkInsertMdash,
			
0))
		    
goto DiskError;
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset + ss - s, MarkRemoveChar,
			
0))
		    
goto DiskError;
		  
ss++;
		
}
	      
	      else if (ss[1] == '-')
		
		{
		  
while (*ss == '-')
		    
		    {
		      
if (ss[1] != '-')
			
			{
			  
if (AddMarkup 
			       (Dataset, LineInfo[0].Offset + ss - s,
				
MarkInsertNdash, 0))
			    
goto DiskError;
			  
if (AddMarkup 
			       (Dataset, LineInfo[0].Offset + ss - s,
				
MarkRemoveChar, 0))
			    
goto DiskError;
			  
ss++;
			
}
		      
		      else
			
			{
			  
if (AddMarkup 
			       (Dataset, LineInfo[0].Offset + ss - s,
				
MarkInsertMdash, 0))
			    
goto DiskError;
			  
if (AddMarkup 
			       (Dataset, LineInfo[0].Offset + ss - s,
				
MarkRemoveChar, 0))
			    
goto DiskError;
			  
if (AddMarkup 
			       (Dataset, LineInfo[0].Offset + ss - s + 1,
				
MarkRemoveChar, 0))
			    
goto DiskError;
			  
ss += 2;
			
}
		    
}
		  
if (*ss != '.' && *ss != '?' && *ss != '!')
		    
		    {
		      
if (AddMarkup 
			   (Dataset, LineInfo[0].Offset + ss - s,
			    
MarkSoftHyphen, 0))
			
goto DiskError;
		    
}
		  
ss--;
		  
continue;
		
}
	    
}
	  
	    // End italics on </I>.
	    if (Italicizing == '<' && *ss == '<' && ss[1] == '/' 
		&&toupper (ss[2]) == 'I' && ss[3] == '>')
	    
	    {
	      
j = 4;
	    
EndItalics:	// Set j before coming here.
	      for (; j > 0; j--)
		
		{
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset + ss - s, MarkRemoveChar,
			
0))
		    
goto DiskError;
		  
if (j > 1)
		    
ss++;
		
}
	      
Italicizing = 0;
	      
if (AddMarkup 
		   (Dataset, LineInfo[0].Offset + ss - s, MarkEndItalics, 0))
		
goto DiskError;
	      
continue;
	    
}
	  
	    // End italicizing when other end-delimiter is encountered.
	    if (Italicizing && *ss == Italicizing)
	    
	    {
	      
j = 1 + (Italicizing == '~');
	      
goto EndItalics;
	    
}
	  
	    // Begin italicizing when <I> encountered.
	    if (!Italicizing && BracketItalicsCount < 5 && *ss == '<' 
		&&toupper (ss[1]) == 'I' && ss[2] == '>')
	    
	    {
	      
j = 3;
	      
Italicizing = 'H';
	    
StartItalics:	// Set j and Italicizing before coming.  
	      for (; j > 0; j--)
		
		{
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset + ss - s, MarkRemoveChar,
			
0))
		    
goto DiskError;
		  
if (j > 1)
		    
ss++;
		
}
	      
if (AddMarkup 
		   (Dataset, LineInfo[0].Offset + ss - s, MarkBeginItalics,
		    0))
		
goto DiskError;
	      
continue;
	    
}
	  
	    // Begin italicizing when _ (or / or <) is encountered at the 
	    // front of a word is encountered.
	    if (!Italicizing && *ss == '<' && BracketItalicsCount > 5)
	    
goto BracketEntry;
	  
if (!Italicizing && (*ss == '_' || *ss == '/' || *ss == '<') 
	       &&!(isspace (ss[1]) || ispunct (ss[1])) && (ss == s 
							   ||
isspace (ss[-1])
							   
 ||ss[-1] ==
							   '\'' 
 ||ss[-1] ==
							   '\"' 
 ||ss[-1] ==
							   '(' 
 ||ss[-1] ==
							   '-'))
	    
	    {
	    
BracketEntry:
j = 1;
	      
if (*ss == '<')
		
		{
		  
if (ss[1] == '/')
		    
BracketItalicsCount = -30000;
		  
		  else if (BracketItalicsCount <= 5 && ss[1] != 0 
			   &&ss[2] == '>')
		    
;
		  
		  else
		    
		    {
		      
Italicizing = '>';
		      
BracketItalicsCount++;
		      
goto StartItalics;
		    
}
		
}
	      
	      else
		
		{
		  
Italicizing = *ss;
		  
goto StartItalics;
		
}
	    
}
	  
	    // Begin italicizing when space~~ is encountered.
	    if (!Italicizing && *ss == '~' && ss[1] == '~' 
		&&!(isspace (ss[2]) || ispunct (ss[2])) && (ss == s 
							    ||
isspace (ss
									[-1])
							    
 ||ss[-1] ==
							    '\'' 
 ||ss[-1] ==
							    '\"' 
 ||ss[-1] ==
							    '(' 
 ||ss[-1] ==
							    '-'))
	    
	    {
	      
j = 2;
	      
Italicizing = '~';
	      
goto StartItalics;
	    
}
	  
	    // Words to auto-italicize.
	    if (!Italicizing)
	    
	    {
	      
for (j = 0; j < NUM_AUTO_ITALICIZED; j++)
		
if (0 == strncmp (ss, Italicized[j], LengthItalicized[j]))
		  
		  {
		    
if (ss > s && !isspace (ss[-1]))
		      
continue;
		    
if (AddMarkup 
			 (Dataset, LineInfo[0].Offset + ss - s,
			  
MarkBeginItalics, 0))
		      
goto DiskError;
		    
ss += LengthItalicized[j] - 1;
		    
if (AddMarkup 
			 (Dataset, LineInfo[0].Offset + ss - s,
			  MarkEndItalics, 
0))
		      
goto DiskError;
		    
break;
		  
}
	      
if (j < NUM_AUTO_ITALICIZED)
		
continue;
	    
}
	  

	    // Detect a hanging single '-' at the end 
	    // of a line.  If the hanging dash is not preceded by whitespace, 
	    // like " - " or " -- " then it is undoubtedly part of 
	    // something like "to-day" and needs to have the trailing space 
	    // removed, or else in html it will render as "to- day".  Of course,
	    // there's also the larger question of whether it should be reduced 
	    // simply to "today",  We'll worry about that one later ...
	    if (isspace (*ss) && ss > &s[1] && ss[-1] == '-'
		&& ss[-2] != '-' 
 &&!LineInfo[1].Empty
		&& !LineInfo[1].BeginsWhite)
	    
	    {
	      
		// Search backward for the first non-dash encountered.
		for (sss = ss - 2; sss >= s && *sss == '-'; sss--);
	      
if (sss >= s && isalpha (*sss))
		
		{
		  
for (sss = ss + 1; isspace (*sss); sss++);
		  
if (*sss == '\0')	// Entire remainder of line blank.
		    {
		      
for (sss = ss; *sss; sss++)
			
			{
			  
if (AddMarkup 
			       (Dataset, LineInfo[0].Offset + sss - s,
				
MarkRemoveChar, 0))
			    
goto DiskError;
			
}
		      
ss = sss - 1;
		      
continue;
		    
}
		
}
	    
}
	  

	    // Detect position where two spaces are needed (namely, after colon
	    // or end-of-sentence).
	    if (isspace (*ss))
	    
	    {
	      
CouldBeName = 1;
	      
sss = ss - 1;
	      
if (sss > s && (*sss == '\"' || *sss == '\''))
		
		{
		  
CouldBeName = 0;
		  
sss--;
		
}
	      
if (sss > s && *sss == ':')
		
		{
		  
if (AddMarkup 
		       (Dataset, LineInfo[0].Offset + ss - s, MarkNbsp, 0))
		    
goto DiskError;
		
}
	      
	      else if (sss > s && (*sss == '.' || *sss == '!' || *sss == '?'))
		
		{
		  
if (*sss != '.')
		    
CouldBeName = 0;
		  
		    // We want to check whether the next character is 
		    // capitalized.
		    for (siv = ss + 1; isspace (*siv); siv++);
		  
if (*siv == '\0')	// Next character is on next line.
		    {
		      
if (!LineInfo[1].CapFirstChar)
			
CouldBeName = 0;
		      
if (LineInfo[1].VerseCap)
			
goto CapitalizedAfterPeriod;
		    
}
		  
		  else		// Next character is on the same line.
		    {
		      
if (*siv == '\"' || *siv == '\'')
			
			{
			  
CouldBeName = 0;
			  
siv++;
			
}
		      
if (*siv >= 'A' && *siv <= 'Z')
			
			{
			
CapitalizedAfterPeriod:
			    // Okay, we know that the next char is capitalized, 
			    // but it could still be part of a construction like
			    // "Mr. Smith".  So we need to detect various
			    // honorifics and eliminate them.  Also, we detect
			    // single-letter capitalized abbreviations, as these
			    // could be initials.
			    if (CouldBeName)
			    
			    {
			      
if (sss > s && sss[-1] >= 'A'
				   && sss[-1] <= 'Z' 
 &&(sss == s + 1
							  ||
							  isspace (sss[-2])))
				
				{
				  
				    // Is something like an initial!  No action 
				    // needed.
				}
			      
			      else
				
				{
				  
for (j = 0;
					
j <
					
sizeof (Honorifics) /
					
sizeof (Honorifics[0]); j++)
				    
				    {
				      
for (siv = sss - 1, sv = Honorifics[j];
					    
siv >= s && *sv; siv--, sv++)
					
if (*siv != *sv)
					  
break;
				      
if (*sv == '\0')
					
break;
				    
}
				  
if (j >=
				       
sizeof (Honorifics) /
				       
sizeof (Honorifics[0]))
				    
CouldBeName = 0;
				
}
			    
}
			  
			    // Finally, we believe that we do need to add the
			    // extra space.
			    if (!CouldBeName)
			    
			    {
			      
if (AddMarkup 
				   (Dataset, LineInfo[0].Offset + ss - s,
				    
MarkNbsp, 0))
				
goto DiskError;
			    
}
			
}
		    
}
		
}
	    
}
	
}
      

	// *** All done with this line! ***
	//DoneThisLine:
	// Advance the LineRecord window.  Note, if it isn't obvious, that 
	// it's okay for the window to move past the end of file, since 
	// it is filled with default blank lines in this case.
	for (j = 0; j < BUFFERED_LINES - 1; j++)
	
BufferedLines[j] = BufferedLines[j + 1];
      
if (1 !=
	   
fread (BufferedLines + j, sizeof (LineRecord), 1,
		   
Dataset->LineFile))
	
BufferedLines[j] = DefaultLine;
    
}
  

    // We're at the end, but may still be in a heading, paragraph, etc.
    if (Italicizing)
    
    {
      
if (AddMarkup (Dataset, ftell (Dataset->InputFile), MarkEndItalics, 0))
	
goto DiskError;
    
}
  
if (InSubtitle)
    
    {
      
if (AddMarkup
	   (Dataset, ftell (Dataset->InputFile), MarkEndSubtitle, 0))
	
goto DiskError;
    
}
  
if (InTable)
    
    {
      
if (AddMarkup (Dataset, ftell (Dataset->InputFile), MarkEndTable, 0))
	
goto DiskError;
    
}
  
if (InParagraph)
    
    {
      
if (AddMarkup 
	   (Dataset, ftell (Dataset->InputFile), MarkEndParagraph, 0))
	
goto DiskError;
    
}
  
if (InBlockquote)
    
    {
      
if (AddMarkup 
	   (Dataset, ftell (Dataset->InputFile), MarkEndBlockquote, 0))
	
goto DiskError;
    
}
  
if (InHeader1)
    
    {
      
if (AddMarkup (Dataset, ftell (Dataset->InputFile), MarkEndHeader1, 0))
	
goto DiskError;
    
}
  

ReturnValue = 0;

Done:
fclose (Dataset->LineFile);
  
AddMarkup (Dataset, -1, 0, 0);	// Flush the buffer.
  return (ReturnValue);

DiskError:
fprintf (stderr, "Disk-write error.\n");
  
ReturnValue = 5;
  
goto Done;

}


