///////////////////////////////////////////////////////////////////////////
/*
  Copyright 2001 Ronald S. Burkey

  This file is part of GutenMark.

  GutenMark is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GutenMark is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GutenMark; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	MatchWordlists.c
  Purpose:	Sets various fields in an existing in-memory wordlist
  		against wordlists/namelists read from the filesystem.
  Mods:		11/21/01 RSB	Began.
  		11/29/01 RSB	Appending to globs, which had been used,
				worked in Linux but not FreeBSD or
				MacOS-X.  Therefore, appending to the
				globs has been eliminated.
                12/01/01 RSB    Globbing doesn't appear at all in
                                mingw32.  Rats!  Fortunately,
                                the DOS-like functions findfirst()
                                and findnext() appear, and they can
                                be used to produce a similar effect.
                                I provide a globlike equivalent in
                                a separate file (winglob.c), but must
                                include winglob.h rather than glob.h.
  
  Note that although the documentation (as of 11/21/01) only refers
  to gzipped wordlists, the wordlists can also be unzipped ASCII
  as well.  This takes a LOT more disk space, but might be faster
  under some circumstances.  (But maybe not!)
*/  
///////////////////////////////////////////////////////////////////////////
  
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifdef WIN32
#include "winglob.h"
#else	/* 
 */
#include <glob.h>
#endif	/* 
 */
#include <zlib.h>
#include "libGutenSpell.h"
  
#define FIRSTGLOB (GLOB_NOSORT | GLOB_TILDE)
//#define NEXTGLOB (GLOB_APPEND | FIRSTGLOB)
#define MAX_WORDLISTS 256
  
// In this list, we keep a marker for each globbed filename,
// indicating if it is supposed to be SPELL_NATIVE or SPELL_FOREIGN 
// language.  If there are more globbed files than entries in
// this table, they are assumed to be foreign.
typedef struct 
{
  
int Count;
   
unsigned char List[MAX_WORDLISTS];
   
char *Names[MAX_WORDLISTS];
 
}
 
NativeList;


//------------------------------------------------------------------------
// Reads and processes an on-disk wordlist.  Returns 0 on success or 
// on file not found.  Returns non-zero only on actual error.
// Note that "gz" file operations (like gzopen rather than fopen)
// are used, because the wordlists are usually gzipped.
// The Type parameter is either SPELL_NATIVE or SPELL_FOREIGN.
//
// The strings read from the wordlist must contain ONLY the characters
// identified by IsWordChar as candidates for being in words, and must
// not begin with punctuation.

static int 
ReadWordlist (Wordlist * Words, const char *Filename,
	      
unsigned char LanguageType) 
{
  
char Full[MAXWORDLENGTH], Normalized[MAXWORDLENGTH],
    
TestFull[MAXWORDLENGTH], TestNormalized[MAXWORDLENGTH], *ss;
  
int i, NumLower, NumUpper, FirstUpper, AnyDiacritical, ReturnValue =
    
1, Matched;
  
unsigned char Capitalization;
  
gzFile fp;
  
fp = gzopen (Filename, "rb");
  
if (fp == NULL)
    
goto Okay;
  

    // The file is open!  Let's read and process it.
    while (Z_NULL != gzgets (fp, Full, sizeof (Full)))
    
    {
      
	// Make sure the string is an actual word (not a comment),
	// and trim off garbage like end of line.
	for (ss = Full; *ss; ss++)
	
if (*ss == '\n')
	  
*ss = 0;
      
if (!Full[0] || Full[0] == '#' || isspace (Full[0]))
	
continue;
      
	// Analyze the capitalization of the word.
	for (NumLower = NumUpper = FirstUpper = AnyDiacritical = 0, ss = Full;
	     
*ss; ss++)
	
	{
	  
i = IsWordChar (*ss);
	  
if (i & WORD_LOWER)
	    
NumLower++;
	  
	  else if (i & WORD_UPPER)
	    
	    {
	      
NumUpper++;
	      
if (!NumLower)
		
FirstUpper = 1;
	    
}
	  
if (i & WORD_DIACRITICAL)
	    
AnyDiacritical = 1;
	
}
      
if (!NumUpper)
	
Capitalization = SPELL_LOWERCASE;
      
      else if (!NumLower)
	
Capitalization = SPELL_UPPERCASE;
      
      else if (NumUpper == 1 && FirstUpper == 1)
	
Capitalization = SPELL_CAPITALIZED;
      
      else
	
Capitalization = SPELL_CUSTOMCAP;
      
	// Compute the normalized form of the word.       
	if (0 == DiacriticalNormalize (Full, Normalized, sizeof (Normalized)))
	
Normalized[0] = 0;
      
	// First, see if this word has been used, as-is.
	i = SearchWordlist (Words, Normalized, Full, &Matched);
      
if (Matched)
	
	{
	  
	    // If the word is already marked, we don't have to do
	    // anything more with it.  But if it hasn't been ...
	    if (!Words->Words[i].WordlistStatus)
	    
Words->Words[i].WordlistStatus = LanguageType | Capitalization;
	  
	    // If foreign, though, we set a flag regardless:
	    if (LanguageType == SPELL_FOREIGN)
	    
Words->Words[i].WordlistStatus |= SPELL_NONNATIVE;
	
}
      
	// If all lower-case, check if the capitalized version was used.  This is
	// necessary to prevent short foreign words from matching it.
	if (Capitalization == SPELL_LOWERCASE)
	
	{
	  
strcpy (TestFull, Full);
	  
strcpy (TestNormalized, Normalized);
	  
TestFull[0] = DiacriticalToupper (TestFull[0]);
	  
TestNormalized[0] = DiacriticalToupper (TestNormalized[0]);
	  
i = SearchWordlist (Words, TestNormalized, TestFull, &Matched);
	  
if (Matched)
	    
	    {
	      
		// If the word is already marked, we don't have to do
		// anything more with it.  But if it hasn't been ...
		if (!Words->Words[i].WordlistStatus)
		
Words->Words[i].WordlistStatus =
		  
LanguageType | Capitalization;
	      
		// If foreign, though, we set a flag regardless:
		if (LanguageType == SPELL_FOREIGN)
		
Words->Words[i].WordlistStatus |= SPELL_NONNATIVE;
	    
}
	
}
      
	// Next, test the same thing, but for the case of the word having
	// been used as all-caps.
	if (Capitalization != SPELL_UPPERCASE)
	
	{
	  
strcpy (TestFull, Full);
	  
strcpy (TestNormalized, Normalized);
	  
DiacriticalStrupr (TestFull);
	  
DiacriticalStrupr (TestNormalized);
	  
i = SearchWordlist (Words, TestNormalized, TestFull, &Matched);
	  
if (Matched)
	    
	    {
	      
		// If the word is already marked, we don't have to do
		// anything more with it.  But if it hasn't been ...
		if (!Words->Words[i].WordlistStatus)
		
Words->Words[i].WordlistStatus =
		  
LanguageType | Capitalization;
	      
		// If foreign, though, we set a flag regardless:
		if (LanguageType == SPELL_FOREIGN)
		
Words->Words[i].WordlistStatus |= SPELL_NONNATIVE;
	    
}
	
}
      
	// The tests above assume a 7-bit to 7-bit or 8-bit to 8-bit
	// match of the word in the etext with the word in the wordlist.
	// Another possibility, though, is that a 7-bit word in the etext
	// corresponds to an 8-bit word in the wordlist.  These tests
	// are more complex, because we're interested not merely in 
	// using them for handling ALL-CAPS italicizing, but also for
	// restoration of diacritical marks. So we also need to check
	// for the merely-capitalized case.
	if (!AnyDiacritical)	// wordlist word not 8-bit anyhow.
	continue;
      
	// Check 7-bit to 8-bit with unchanged capitalization.
	i = SearchWordlist (Words, Normalized, Normalized, &Matched);
      
if (Matched)
	
	{
	  
	    // If the word is already marked, we don't have to do
	    // anything more with it.  But if it hasn't been ...
	    if (!Words->Words[i].WordlistStatus)
	    
	    {
	      
if (NULL != 
		   (Words->Words[i].Match = AllocSpellString (Words, Full)))
		
Words->Words[i].WordlistStatus =
		  
SPELL_NORMALIZED | LanguageType | Capitalization;
	    
}
	  
	    // If foreign, though, we set a flag regardless:
	    if (LanguageType == SPELL_FOREIGN)
	    
Words->Words[i].WordlistStatus |= SPELL_NONNATIVE;
	
}
      
	// Check 7-bit to 8-bit with initial capitalization.
	Normalized[0] = DiacriticalToupper (Normalized[0]);
      
i = SearchWordlist (Words, Normalized, Normalized, &Matched);
      
if (Matched)
	
	{
	  
	    // If the word is already marked, we don't have to do
	    // anything more with it.  But if it hasn't been ...
	    if (!Words->Words[i].WordlistStatus)
	    
	    {
	      
if (NULL != 
		   (Words->Words[i].Match = AllocSpellString (Words, Full)))
		
Words->Words[i].WordlistStatus =
		  
SPELL_NORMALIZED | LanguageType | Capitalization;
	    
}
	  
	    // If foreign, though, we set a flag regardless:
	    if (LanguageType == SPELL_FOREIGN)
	    
Words->Words[i].WordlistStatus |= SPELL_NONNATIVE;
	
}
      
	// Check 7-bit to 8-bit with all caps.
	i = SearchWordlist (Words, TestNormalized, TestNormalized, &Matched);
      
if (Matched)
	
	{
	  
	    // If the word is already marked, we don't have to do
	    // anything more with it.  But if it hasn't been ...
	    if (!Words->Words[i].WordlistStatus)
	    
	    {
	      
if (NULL != 
		   (Words->Words[i].Match = AllocSpellString (Words, Full)))
		
Words->Words[i].WordlistStatus =
		  
SPELL_NORMALIZED | LanguageType | Capitalization;
	    
}
	  
	    // If foreign, though, we set a flag regardless:
	    if (LanguageType == SPELL_FOREIGN)
	    
Words->Words[i].WordlistStatus |= SPELL_NONNATIVE;
	
}
    
}


Okay:
ReturnValue = 0;
  
//Done:
    if (fp != NULL)
    
gzclose (fp);
  
return (ReturnValue);

}



//------------------------------------------------------------------------
// Update the NativeList array after globbing.  The glob_t contains
// the globbed filenames derived from an entry in the configuration
// file.

void 
Mark (NativeList * List, glob_t * Glob, char Type) 
{
  
int i, j;
  
for (i = 0, j = List->Count; i < Glob->gl_pathc && j < MAX_WORDLISTS;
	
i++, j++)
    
    {
      
List->List[j] = Type;
      
List->Names[j] = (char *) calloc (1, strlen (Glob->gl_pathv[i]) + 1);
      
if (List->Names[j] == NULL)
	
break;
      
strcpy (List->Names[j], Glob->gl_pathv[i]);
    
}
  
List->Count = j;

}



//------------------------------------------------------------------------
// libGutenSpell operates in reverse from the way a normal spell-checker
// works.  A normal spell-checker takes a selected word and tries to 
// find it in on-disk wordlists.  This forces the on-disk wordlists to
// be highly organized for fast searching.  With libGutenSpell, however,
// it is the in-memory words that are highly organized, and the on-disk
// wordlists need no particular organization (i.e., they don't have to
// be sorted), because EVERY word in the on-disk wordlists is going to
// be read and checked against the in-memory list.  This has the advantage
// of allowing the on-disk wordlists to be highly compressed.  In fact,
// they are read with zlib rather than stdio.
//
// The result of running this function is merely to set certain fields
// (mostly flags) in the in-memory wordlist, representing properties of 
// the various words:  i.e., whether they were found or not, whether
// native language or foreign, whether 7-bit ASCII or 8-bit ASCII, etc.
//
// The wordlists to be used are taken from the GutenMark.cfg file.
// They may contain wildcards or other regular-expression-type stuff,
// and so the full list of wordlists needs to be constructed by globbing,
// and this is done before any of the wordlists are actually read.
//
// Returns zero on success.

int 
MatchWordlists (FILE * LogFile, Wordlist * Words, const char *Language) 
{
  
NativeList Natives =
  {
  0};
  
glob_t Glob;
  
char s[128], *ss;
  
FILE * cfg;
  
int i, j, ReturnValue, Found = 0;
  
Glob.gl_offs = Glob.gl_pathc = 0;
  
Glob.gl_pathv = NULL;
  
cfg = fopen ("GutenMark.cfg", "r");
  
    // Loop on the lines in the configuration file.  (If the configuration
    // file wasn't found, we simply do all of the namelists, followed
    // by all of the wordlists.)
    if (cfg == NULL)
    
    {
    
Default:
sprintf (s, "%s.names.gz", Language);
      
ReturnValue = glob (s, FIRSTGLOB, NULL, &Glob);
      
Mark (&Natives, &Glob, SPELL_NATIVE);
      
ReturnValue |= glob ("*.names.gz", FIRSTGLOB, NULL, &Glob);
      
Mark (&Natives, &Glob, SPELL_NATIVE);
      
sprintf (s, "%s.words.gz", Language);
      
ReturnValue |= glob (s, FIRSTGLOB, NULL, &Glob);
      
Mark (&Natives, &Glob, SPELL_NATIVE);
      
ReturnValue |= glob ("*.words.gz", FIRSTGLOB, NULL, &Glob);
      
Mark (&Natives, &Glob, SPELL_FOREIGN);
      
if (ReturnValue)
	
goto Done;
    
}
  
  else
    
    {
      
	// Search the cfg file for the appropriate language profile.
	// These are headed by lines like "[language]".
	while (NULL != fgets (s, sizeof (s) - 1, cfg))
	
if (s[0] == '[')
	  
	  {
	    
for (ss = &s[1]; *ss; ss++)
	      
if (*ss == ']')
		
break;
	    
if (*ss == ']')
	      
	      {
		
*ss = 0;
		
		  // Language profile found?
		  if (!strcasecmp (&s[1], Language))
		  
		  {
		    
Found = 1;
		    
		      // We now use all lines between now and the
		      // next profile (or the end of file) beginning
		      // with "native=" or "foreign=".
		      while (NULL != fgets (s, sizeof (s) - 1, cfg))
		      
		      {
			
if (s[0] == '[')
			  
break;
			
for (ss = s; *ss; ss++)
			  
if (*ss == '\n')
			    
*ss = 0;
			
if (!strncasecmp (s, "native=", 7))
			  
			  {
			    
ReturnValue =
			      
glob (&s[7], FIRSTGLOB, NULL, &Glob);
			    
Mark (&Natives, &Glob, SPELL_NATIVE);
			    
if (ReturnValue)
			      
break;
			  
}
			
			else if (!strncasecmp (s, "foreign=", 8))
			  
			  {
			    
ReturnValue =
			      
glob (&s[8], FIRSTGLOB, NULL, &Glob);
			    
Mark (&Natives, &Glob, SPELL_FOREIGN);
			    
if (ReturnValue)
			      
break;
			  
}
		      
}
		    
break;
		  
}
	      
}
	  
}
      
fclose (cfg);
      
if (!Found)
	
goto Default;
    
}
  
    // Now the list of all wordslists and namelists has been totally
    // expanded, though possibly with duplicates.  We therefore 
    // process, them one-by-one.  
    for (i = 0; i < Natives.Count; i++)
    
    {
      
	// Make sure one isn't a duplicate.
	for (j = 0; j < i; j++)
	
if (!strcmp (Natives.Names[i], Natives.Names[j]))
	  
break;
      
if (j == i)
	
	{
	  
if (LogFile != NULL)
	    
fprintf (LogFile, "Wordlist = %s\n", Natives.Names[i]);
	  
fprintf (stderr, "Checking wordlist/namelist %s\n",
		    
Natives.Names[i]);
	  
	    // Not a duplicate.  Process it!
	    ReturnValue =
	    
ReadWordlist (Words, Natives.Names[i], Natives.List[i]);
	  
if (ReturnValue)
	    
goto Done;
	
}
      
      else
	
	{
	  
if (LogFile != NULL)
	    
fprintf (LogFile, "Duplicate wordlist = %s\n", Natives.Names[i]);
	
}
    
}
  
    // All done!  
    ReturnValue = 0;

Done:
    // For some reason, the following can cause a segmentation fault.
    //globfree (&Glob);
    return (ReturnValue);

}


